@@ -3,10 +3,10 @@
 Test stored procedure usage.
 """
 
+from contextlib import contextmanager
 import decimal
 import datetime
 import os
-import sys
 import unittest
 
 import pymssql
@@ -16,9 +16,46 @@ import pytest
 
 from .helpers import mssqlconn, pymssqlconn
 
+
+@contextmanager
+def stored_proc(mssqlcon, dbtype):
+
+    dbtype = dbtype.lower()
+    identifier = dbtype
+    if dbtype == 'decimal':
+        identifier = 'decimal(6, 5)'
+    elif dbtype == 'numeric':
+        identifier = 'numeric(6, 5)'
+    elif dbtype == 'time':
+        identifier = 'time(7)'
+    elif dbtype == 'datetimeoffset':
+        identifier = 'datetimeoffset(7)'
+
+    name = f"pymssqlTest_{dbtype}"
+
+    cmd = f"""
+        CREATE PROCEDURE [dbo].[{name}]
+            @inp {identifier},
+            @out {identifier} output
+        AS
+        BEGIN
+            SET @out = @inp;
+            RETURN 0;
+        END
+    """
+    mssqlcon.execute_non_query(cmd)
+
+    try:
+        yield name
+    finally:
+        # Code to release resource, e.g.:
+        mssqlcon.execute_non_query(f'DROP PROCEDURE [dbo].[{name}]')
+
+
 FIXED_TYPES = (
     'BigInt',
     'Bit',
+    'Date',
     'DateTime',
     'Decimal',
     'Int',
@@ -76,6 +113,13 @@ class TestFixedTypeConversion(unittest.TestCase):
         self.pymssql.close()
         self.mssql.close()
 
+    def testBigIntNullIn(self):
+        input = None
+        proc = self.mssql.init_procedure('pymssqlTestBigInt')
+        proc.bind(input, _mssql.SQLINT8, '@ibigint', null=True)
+        proc.bind(None, _mssql.SQLINT8, '@obigint', output=True)
+        proc.execute()
+        self.assertEqual(input, proc.parameters['@obigint'])
 
     def testBigInt(self):
         input = 123456789
@@ -87,19 +131,13 @@ class TestFixedTypeConversion(unittest.TestCase):
 
     def testBigIntPymssql(self):
         """Same as testBigInt above but from pymssql. Uses pymssql.output class."""
-
-        if sys.version_info >= (3, ):
-            py_type = int
-        else:
-            py_type = long
-
         in_val = 123456789
         cursor = self.pymssql.cursor()
-        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(py_type)])
+        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(int)])
         self.assertEqual(in_val, retval[1])
 
         in_val = 2147483647
-        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(py_type)])
+        retval = cursor.callproc('pymssqlTestBigInt', [in_val, pymssql.output(int)])
         self.assertEqual(in_val, retval[1])
 
     def testBit(self):
@@ -238,21 +276,110 @@ class TestFixedTypeConversion(unittest.TestCase):
         assert abs(input - proc.parameters['@ofloat']) < 0.00001
 
 
+@pytest.mark.mssql_server_required
+class TestDateTimeConversion:
+
+    def test_date(self, mssql_conn, subtests):
+        with stored_proc(mssql_conn, 'date') as proc_name:
+            for input in (
+                datetime.date(1, 1, 1), # min supported
+                datetime.date(90, 8, 27), # GH #454
+                datetime.date(906, 8, 27), # GH #454
+                datetime.date(2009, 8, 27),
+                datetime.date(9999, 12, 31), # max supported
+                ):
+                with subtests.test(date=input):
+                    proc = mssql_conn.init_procedure(proc_name)
+                    proc.bind(input, _mssql.SQLDATE, '@inp')
+                    proc.bind(None, _mssql.SQLDATE, '@out', output=True)
+                    res = proc.execute()
+                    assert res == 0
+                    assert input == proc.parameters['@out']
+
+    def test_time(self, mssql_conn, subtests):
+        with stored_proc(mssql_conn, 'time') as proc_name:
+            for input in (
+                datetime.time(0, 0, 0), # min supported
+                datetime.time(23, 59, 59, 0),
+                datetime.time(23, 59, 59, 999999), # max supported
+                ):
+                with subtests.test(time=input):
+                    proc = mssql_conn.init_procedure(proc_name)
+                    proc.bind(input, _mssql.SQLTIME, '@inp')
+                    proc.bind(None, _mssql.SQLTIME, '@out', output=True)
+                    res = proc.execute()
+                    assert res == 0
+                    assert input == proc.parameters['@out']
+
+    def test_datetime(self, mssql_conn, subtests):
+        with stored_proc(mssql_conn, 'datetime') as proc_name:
+            for input in (
+                datetime.datetime(1753, 1, 1), # min supported
+                datetime.datetime(2009, 8, 27),
+                datetime.datetime(2009, 8, 27, 23, 59, 59, 997000),
+                datetime.datetime(9999, 12, 31, 23, 59, 59, 997000), # max supported
+                ):
+                with subtests.test(datetime=input):
+                    proc = mssql_conn.init_procedure(proc_name)
+                    proc.bind(input, _mssql.SQLDATETIME, '@inp')
+                    proc.bind(None, _mssql.SQLDATETIME, '@out', output=True)
+                    res = proc.execute()
+                    assert res == 0
+                    assert input == proc.parameters['@out']
+
+    def test_datetime2(self, mssql_conn, subtests):
+        with stored_proc(mssql_conn, 'datetime2') as proc_name:
+            for input in (
+                datetime.datetime(1, 1, 1), # min supported
+                datetime.datetime(90, 8, 27),
+                datetime.datetime(906, 8, 27),
+                datetime.datetime(2009, 8, 27),
+                datetime.datetime(2009, 8, 27, 23, 59, 59, 999999),
+                datetime.datetime(9999, 12, 31, 23, 59, 59, 999999), # max supported
+                ):
+                with subtests.test(datetime=input):
+                    proc = mssql_conn.init_procedure(proc_name)
+                    proc.bind(input, _mssql.SQLDATETIME2, '@inp')
+                    proc.bind(None, _mssql.SQLDATETIME2, '@out', output=True)
+                    res = proc.execute()
+                    assert res == 0
+                    assert input == proc.parameters['@out']
+
+    def test_datetimeoffset(self, mssql_conn, subtests):
+        with stored_proc(mssql_conn, 'datetimeoffset') as proc_name:
+            for h in range(-12,13,1):
+                for m in range(0, 61, 15):
+                    tz = datetime.timezone(datetime.timedelta(hours=h, minutes=m))
+                    for input in (
+                        datetime.datetime(10, 1, 2, tzinfo=tz), # min supported
+                        datetime.datetime(2999, 12, 31, 23, 59, 59, 999999, tzinfo=tz), # max supported
+                        ):
+                        with subtests.test(datetime=input):
+                            proc = mssql_conn.init_procedure(proc_name)
+                            proc.bind(input, _mssql.SQLDATETIMEOFFSET, '@inp')
+                            proc.bind(None, _mssql.SQLDATETIMEOFFSET, '@out', output=True)
+                            res = proc.execute()
+                            assert res == 0
+                            assert input == proc.parameters['@out']
+
+
 @pytest.mark.mssql_server_required
 class TestCallProcFancy(unittest.TestCase):
-    # "Fancy" because we test some exotic cases like passing None or Unicode
+    """
+    "Fancy" because we test some exotic cases like passing None or Unicode
     # strings to a called procedure
+    """
 
     def setUp(self):
         self.pymssql = pymssqlconn()
         cursor = self.pymssql.cursor()
 
-        sql = u"""
+        sql = """
         CREATE PROCEDURE [dbo].[someProcWithOneParam]
-        	@some_arg NVARCHAR(64)
+            @some_arg NVARCHAR(64)
         AS
         BEGIN
-        	SELECT @some_arg + N'!',
+            SELECT @some_arg + N'!',
                    N'%(str1)s ' + @some_arg + N' %(str2)s'
         END
         """ % {
@@ -421,27 +548,20 @@ class TestStringTypeConversion(unittest.TestCase):
             self.assertEqual(input, proc.parameters['@ovarbinary'])
             self.assertEqual(output_type, type(proc.parameters['@ovarbinary']))
 
-        if sys.version_info[0] == 3:
-            check_conversion(bytes(b'\xDE\xAD\xBE\xEF'), bytes)
-            check_conversion(bytearray(b'\xDE\xAD\xBE\xEF'), bytes)
-            with pytest.raises(TypeError) as exc_info:
-                check_conversion('FOO', bytes)
-                assert 'value can only be bytes or bytearray' == str(exc_info.value)
-        else:
-            check_conversion(b'\xDE\xAD\xBE\xEF', str)
-            check_conversion(bytes(b'\xDE\xAD\xBE\xEF'), str)
-            check_conversion(bytearray(b'\xDE\xAD\xBE\xEF'), str)
-            with pytest.raises(TypeError) as exc_info:
-                check_conversion(unicode('Foo'), str)
-                assert 'value can only be str or bytearray' == str(exc_info.value)
+        check_conversion(bytes(b'\xDE\xAD\xBE\xEF'), bytes)
+        check_conversion(bytearray(b'\xDE\xAD\xBE\xEF'), bytes)
+        with pytest.raises(TypeError) as exc_info:
+            check_conversion('FOO', bytes)
+            assert 'value can only be bytes or bytearray' == str(exc_info.value)
 
 
 @pytest.mark.mssql_server_required
 class TestFloatTypeConversion(unittest.TestCase):
+
     def setUp(self):
         self.mssql = mssqlconn()
         self.pymssql = pymssqlconn()
-        cursor = self.pymssql.cursor()
+        self.cursor = self.pymssql.cursor()
 
         self.mssql.execute_non_query("""
             CREATE PROCEDURE [dbo].[pymssqlRealTest]
@@ -464,39 +584,82 @@ class TestFloatTypeConversion(unittest.TestCase):
         )
 
     def tearDown(self):
-        cursor = self.pymssql.cursor()
         self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlRealTest]')
         self.mssql.execute_non_query('DROP PROCEDURE [dbo].[pymssqlFloatTest]')
+        self.mssql.close()
+        self.cursor.close()
         self.pymssql.close()
 
-    def testReal(self):
-        cursor = self.pymssql.cursor()
-        cursor.callproc(
-            'pymssqlRealTest',
-            (0.5,))
-
-        # TODO: Use the solution we implement once #134 gets fixed
-        a = next(cursor)
+    def test_real(self):
+        self.cursor.callproc('pymssqlRealTest', (0.5,))
+        a = next(self.cursor)
+        assert len(a) == 1
         assert abs(a[0] - 0.5) < 0.000001
 
-    def testFloat8(self):
-        cursor = self.pymssql.cursor()
-        cursor.callproc(
-            'pymssqlFloatTest',
-            (5.44451787074e+39,))
-
-        # TODO: Use the solution we implement once #134 gets fixed
-        a = next(cursor)
+    def test_float8(self):
+        self.cursor.callproc('pymssqlFloatTest', (5.44451787074e+39,))
+        a = next(self.cursor)
+        assert len(a) == 1
         assert abs(a[0] - 5.44451787074e+39) < 0.000001
 
+
+@pytest.mark.mssql_server_required
+class TestFetch(unittest.TestCase):
+
+    def setUp(self):
+        self.pymssql = pymssqlconn()
+        self.cursor = self.pymssql.cursor()
+
+        self.cursor.execute("""
+            CREATE PROCEDURE [dbo].[pymssqlTest]
+                @inparam real
+            AS
+            BEGIN
+                SELECT @inparam AS outparam
+            END
+            """
+        )
+
+    def tearDown(self):
+        self.cursor.execute('DROP PROCEDURE [dbo].[pymssqlTest]')
+        self.cursor.close()
+        self.pymssql.close()
+
+    def test_fetchall(self):
+        self.cursor.callproc('pymssqlTest', (0.5,))
+        a = self.cursor.fetchall()
+        assert len(a) == 1
+        assert len(a[0]) == 1
+        assert abs(a[0][0] - 0.5) < 0.000001
+
+    def test_fetchone(self):
+        self.cursor.callproc('pymssqlTest', (0.5,))
+        a = self.cursor.fetchone()
+        assert len(a) == 1
+        assert abs(a[0] - 0.5) < 0.000001
+
+    def test_fetchmany(self):
+        self.cursor.callproc('pymssqlTest', (0.5,))
+        a = self.cursor.fetchmany()
+        assert len(a) == 1
+        assert len(a[0]) == 1
+        assert abs(a[0][0] - 0.5) < 0.000001
+
+    def test_real_next(self):
+        self.cursor.callproc('pymssqlTest', (0.5,))
+        a = next(self.cursor)
+        assert len(a) == 1
+        assert abs(a[0] - 0.5) < 0.000001
+
+
 @pytest.mark.mssql_server_required
 class TestErrorInSP(unittest.TestCase):
 
     def setUp(self):
         self.pymssql = pymssqlconn()
-        cursor = self.pymssql.cursor()
+        self.cursor = self.pymssql.cursor()
 
-        sql = u"""
+        sql = """
         CREATE PROCEDURE [dbo].[SPThatRaisesAnError]
         AS
         BEGIN
@@ -507,27 +670,29 @@ class TestErrorInSP(unittest.TestCase):
             RETURN
         END
         """
-        cursor.execute(sql)
+        self.cursor.execute(sql)
 
     def tearDown(self):
-        cursor = self.pymssql.cursor()
-        cursor.execute('DROP PROCEDURE [dbo].[SPThatRaisesAnError]')
+        self.cursor.execute('DROP PROCEDURE [dbo].[SPThatRaisesAnError]')
+        self.cursor.close()
         self.pymssql.close()
 
     def test_tsql_to_python_exception_translation(self):
-        """An error raised by a SP is translated to a PEP-249-dictated, pymssql layer exception."""
-        # See https://github.com/pymssql/pymssql/issues/61
-        cursor = self.pymssql.cursor()
+        """
+        An error raised by a SP is translated to a PEP-249-dictated,
+        pymssql layer exception.
+        See https://github.com/pymssql/pymssql/issues/61
+        """
         # Must raise an exception
-        self.assertRaises(Exception, cursor.callproc, 'SPThatRaisesAnError')
+        self.assertRaises(Exception, self.cursor.callproc, 'SPThatRaisesAnError')
         # Must be a PEP-249 exception, not a _mssql-layer one
         try:
-            cursor.callproc('SPThatRaisesAnError')
+            self.cursor.callproc('SPThatRaisesAnError')
         except Exception as e:
             self.assertTrue(isinstance(e,  pymssql.Error))
         # Must be a DatabaseError exception
         try:
-            cursor.callproc('SPThatRaisesAnError')
+            self.cursor.callproc('SPThatRaisesAnError')
         except Exception as e:
             self.assertTrue(isinstance(e,  pymssql.DatabaseError))
 
@@ -555,7 +720,7 @@ class TestSPWithQueryResult(unittest.TestCase):
         self.pymssql.close()
         self.mssql.close()
 
-    def testPymssql(self):
+    def test_pymssql(self):
         cursor = self.pymssql.cursor()
         cursor.callproc(
             self.SP_NAME,
@@ -572,3 +737,86 @@ class TestSPWithQueryResult(unittest.TestCase):
 
         for row_dict in self.mssql:
             self.assertEqual(row_dict, {0: 'hello!', 1: 'hello!!'})
+
+
+@pytest.mark.mssql_server_required
+class TestMultipleDataSetsInResult(unittest.TestCase):
+    """
+    Test for https://github.com/pymssql/pymssql/pull/134
+    See also https://gist.github.com/msabramo/6747703
+    """
+    table_name = 'callproc_demo_tbl'
+    proc_name = 'callproc_get_2_resultsets'
+
+    def setUp(self):
+        self.pymssql = pymssqlconn()
+        self.cursor = self.pymssql.cursor()
+        self.cursor.execute(f"IF OBJECT_ID('{self.table_name}', 'U') IS NOT NULL DROP TABLE {self.table_name}")
+        self.cursor.execute(f"CREATE TABLE {self.table_name} (name VARCHAR(30))")
+        self.cursor.execute(f"INSERT INTO {self.table_name} VALUES ('Tom'), ('Dick'), ('Harry')")
+        self.cursor.execute(f"""
+        CREATE PROCEDURE [dbo].[{self.proc_name}] AS
+        BEGIN
+            SET NOCOUNT ON
+            SELECT * FROM {self.table_name} WHERE name = 'not here';
+            SELECT * FROM {self.table_name};
+        END
+        """)
+
+    def tearDown(self):
+        self.cursor.execute(f'DROP PROCEDURE [dbo].[{self.proc_name}]')
+        self.cursor.close()
+        self.pymssql.close()
+
+    def test_exec(self):
+        self.cursor.execute(f'EXEC {self.proc_name}')
+        res = self.cursor.fetchall()
+        self.assertEqual(res, [])
+        res = self.cursor.fetchall()
+        self.assertEqual(res, [('Tom',), ('Dick',), ('Harry',)])
+
+    @pytest.mark.xfail(strict=True)
+    def test_callproc(self):
+        self.cursor.callproc(self.proc_name)
+        res = self.cursor.fetchall()
+        self.assertEqual(res, [])
+        res = self.cursor.fetchall()
+        self.assertEqual(res, [('Tom',), ('Dick',), ('Harry',)])
+
+
+@pytest.mark.mssql_server_required
+class TestSPWithNULLResult(unittest.TestCase):
+    """
+    GH: #441
+    """
+
+    SP_NAME = 'SPWithANullResult'
+
+    def setUp(self):
+        self.mssql = mssqlconn()
+        self.pymssql = pymssqlconn()
+
+        self.mssql.execute_non_query("""
+        CREATE PROCEDURE [dbo].[%(spname)s]
+                @int_arg int output,
+                @varchar_arg varchar output
+        AS
+        BEGIN
+                SET @int_arg = null
+                SET @varchar_arg = null
+        END
+        """ % {'spname': self.SP_NAME})
+
+    def tearDown(self):
+        self.mssql.execute_non_query('DROP PROCEDURE [dbo].[%(spname)s]' % {'spname': self.SP_NAME})
+        self.pymssql.close()
+        self.mssql.close()
+
+    def testNull(self):
+        proc = self.mssql.init_procedure(self.SP_NAME)
+        proc.bind(None, _mssql.SQLINT8, '@int_arg', output=True)
+        proc.bind(None, _mssql.SQLVARCHAR, '@varchar_arg', output=True)
+        proc.execute()
+        self.assertIsNone(proc.parameters['@int_arg'])
+        self.assertIsNone(proc.parameters['@varchar_arg'])
+
